#ifndef NEWSBOAT_ITEMVIEW_H_
#define NEWSBOAT_ITEMVIEW_H_

static char itemview_str[] = "" "!vbox\n"
"  @style_normal[background]:\n"
"  @info#style_normal[info]:\n"
"  @info#style_key_normal[hint-key]:\n"
"  @info#style_comma_normal[hint-keys-delimiter]:\n"
"  @info#style_colon_normal[hint-separator]:\n"
"  @info#style_desc_normal[hint-description]:\n"
"  @title#style_normal[title]:\n"
"  @bind_up[bind_up]:\n"
"  @bind_down[bind_down]:\n"
"  @bind_page_up[bind_page_up]:\n"
"  @bind_page_down[bind_page_down]:\n"
"  @bind_home[bind_home]:\n"
"  @bind_end[bind_end]:\n"
"  @on_TAB:\"TAB\"\n"
"  label#title[title]\n"
"    text[head]:\"Article '#'\"\n"
"    .expand:h\n"
"    .display[showtitle]:1\n"
"  textview[article]\n"
"    style_normal[article]:\n"
"    style_end[end-of-text-marker]:\n"
"    .expand:vh\n"
"    offset[article_offset]:0\n"
"    richtext:1\n"
"  vbox[hints]\n"
"    .expand:0\n"
"    .display[showhint]:1\n"
"    hbox\n"
"      .expand:0\n"
"      label#info\n"
"        text[help]:\"\"\n"
"        richtext:1\n"
"        .expand:h\n"
"      label#info\n"
"        text[percent]:\"\"\n"
"        .expand:0\n"
"        .width[percentwidth]:6\n"
"  hbox[lastline]\n"
"    .expand:0\n"
"    label\n"
"      text[msg]:\"\"\n"
"      .expand:h\n"
"      .display[show_msg]:1\n"
"    label\n"
"      .expand:0\n"
"      text[qna_prompt]:\"\"\n"
"      .display[show_qna_prompt]:0\n"
"    input[qnainput]\n"
"      modal:1\n"
"      .expand:h\n"
"      @bind_home:**\n"
"      @bind_end:**\n"
"      text[qna_value]:\"\"\n"
"      pos[qna_value_pos]:0\n"
"      .display[show_qna_input]:0\n"
;

#endif /* NEWSBOAT_ITEMVIEW_H_ */
