#ifndef NEWSBOAT_EMPTY_H_
#define NEWSBOAT_EMPTY_H_

static char empty_str[] = "" "vbox\n"
"  label#title[title]\n"
"    text[head]:\"empty formaction\"\n"
"    .expand:h\n"
"    .display[showtitle]:1\n"
"  vbox[hints]\n"
"    .expand:0\n"
"    .display[showhint]:1\n"
"  hbox[lastline]\n"
"    .expand:0\n"
"    label\n"
"      text[msg]:\"\"\n"
"      .expand:h\n"
"      .display[show_msg]:1\n"
"    label\n"
"      .expand:0\n"
"      text[qna_prompt]:\"\"\n"
"      .display[show_qna_prompt]:0\n"
"    input[qnainput]\n"
"      modal:1\n"
"      .expand:h\n"
"      @bind_home:**\n"
"      @bind_end:**\n"
"      text[qna_value]:\"\"\n"
"      pos[qna_value_pos]:0\n"
"      .display[show_qna_input]:0\n"
;

#endif /* NEWSBOAT_EMPTY_H_ */
